from typing import Any, Dict, Tuple

import oml.models.vit_unicom.external.vision_transformer as unicom  # type: ignore
from oml.transforms.images.albumentations import (
    get_augs_albu,
    get_normalisation_albu,
    get_normalisation_resize_albu,
    get_normalisation_resize_albu_clip,
)
from oml.transforms.images.torchvision import (
    get_augs_hypvit,
    get_augs_torch,
    get_normalisation_resize_hypvit,
    get_normalisation_resize_torch,
    get_normalisation_torch,
)
from oml.transforms.images.utils import TTransforms, get_im_reader_for_transforms
from oml.utils.images.images import TImReader
from oml.utils.misc import TCfg, dictconfig_to_dict

TRANSFORMS_ALBU = {
    "augs_albu": get_augs_albu,
    "norm_albu": get_normalisation_albu,
    "norm_resize_albu": get_normalisation_resize_albu,
    "norm_resize_albu_clip": get_normalisation_resize_albu_clip,
}

TRANSFORMS_TORCH = {
    "augs_torch": get_augs_torch,
    "norm_torch": get_normalisation_torch,
    "norm_resize_torch": get_normalisation_resize_torch,
    "augs_hypvit_torch": get_augs_hypvit,
    "norm_resize_hypvit_torch": get_normalisation_resize_hypvit,
    "unicom_transforms": unicom.transform,  # type: ignore
}

TRANSFORMS_REGISTRY = {**TRANSFORMS_ALBU, **TRANSFORMS_TORCH}


def get_transforms(name: str, **kwargs: Dict[str, Any]) -> TTransforms:
    augs = TRANSFORMS_REGISTRY[name](**kwargs)  # type: ignore
    return augs


def get_transforms_by_cfg(cfg: TCfg) -> TTransforms:
    cfg = dictconfig_to_dict(cfg)
    return get_transforms(name=cfg["name"], **cfg["args"])


TRANSFORMS_FOR_PRETRAINED = {
    "resnet50_moco_v2": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet18_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet34_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet50_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet101_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet152_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "vitb8_dino": get_normalisation_resize_torch(im_size=224),
    "vitb16_dino": get_normalisation_resize_torch(im_size=224),
    "vits8_dino": get_normalisation_resize_torch(im_size=224),
    "vits16_dino": get_normalisation_resize_torch(im_size=224),
    "sber_vitb32_224": get_normalisation_resize_albu_clip(im_size=224),
    "sber_vitb16_224": get_normalisation_resize_albu_clip(im_size=224),
    "sber_vitl14_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitb32_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitb16_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitl14_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitl14_336": get_normalisation_resize_albu_clip(im_size=224),
    "vits16_inshop": get_normalisation_resize_hypvit(im_size=224, crop_size=224),
    "vits16_sop": get_normalisation_resize_hypvit(im_size=224, crop_size=224),
    "vits16_cars": get_normalisation_resize_albu(im_size=224),
    "vits16_cub": get_normalisation_resize_albu(im_size=224),
    "vits16_224_mlp_384_inshop": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "vitb32_unicom": unicom.transform(224),  # type: ignore
    "vitb16_unicom": unicom.transform(224),  # type: ignore
    "vitl14_unicom": unicom.transform(224),  # type: ignore
    "vitl14_336px_unicom": unicom.transform(336),  # type: ignore
}


def get_transforms_for_pretrained(weights: str) -> Tuple[TTransforms, TImReader]:
    transforms = TRANSFORMS_FOR_PRETRAINED[weights]
    im_reader = get_im_reader_for_transforms(transforms)
    return transforms, im_reader


__all__ = [
    "TRANSFORMS_TORCH",
    "TRANSFORMS_ALBU",
    "TRANSFORMS_REGISTRY",
    "get_transforms",
    "get_transforms_by_cfg",
    "get_transforms_for_pretrained",
]



from typing import Any, Dict, Union

from pytorch_lightning.loggers import (
    LightningLoggerBase,
    NeptuneLogger,
    TensorBoardLogger,
    WandbLogger,
)

from oml.const import TCfg
from oml.utils.misc import dictconfig_to_dict

LOGGERS_REGISTRY = {"wandb": WandbLogger, "neptune": NeptuneLogger, "tensorboard": TensorBoardLogger}


def get_logger(name: str, **kwargs: Dict[str, Any]) -> LightningLoggerBase:
    return LOGGERS_REGISTRY[name](**kwargs)


def get_logger_by_cfg(cfg: TCfg) -> Union[bool, LightningLoggerBase]:
    cfg = dictconfig_to_dict(cfg)
    logger = get_logger(cfg["name"], **cfg["args"])
    return logger


__all__ = ["LOGGERS_REGISTRY", "get_logger", "get_logger_by_cfg"]



import multiprocessing as mp
from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Callable, List, Optional, Union

from tqdm import tqdm

try:
    from albumentations.augmentations.functional import pad
except (AttributeError, ModuleNotFoundError, ImportError):
    from albumentations.augmentations.geometric.functional import pad

import cv2
import numpy as np
import PIL
import torch
from PIL import Image
from PIL.Image import Image as TPILImage

from oml.const import PAD_COLOR, TColor

TImage = Union[PIL.Image.Image, np.ndarray]
TImReader = Callable[[Union[Path, str, bytes]], TImage]


def tensor_to_numpy_image(img: torch.Tensor) -> np.ndarray:
    img = img.detach().cpu()
    img = img.permute(1, 2, 0).numpy().copy()

    img -= img.min()
    img /= img.max()
    img *= 255

    img = img.astype(np.uint8)

    return img


def imread_cv2(im_src: Union[Path, str, bytes]) -> np.ndarray:
    if isinstance(im_src, (Path, str)):
        image = cv2.imread(str(im_src), cv2.IMREAD_UNCHANGED)
    elif isinstance(im_src, bytes):
        image_raw = np.frombuffer(im_src, np.uint8)
        image = cv2.imdecode(image_raw, cv2.IMREAD_UNCHANGED)
    else:
        raise TypeError("Unsupported type")

    return cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)


def imread_pillow(im_src: Union[Path, str, bytes]) -> TPILImage:
    if isinstance(im_src, (Path, str)):
        image = Image.open(im_src)
    elif isinstance(im_src, bytes):
        image = Image.open(BytesIO(im_src))
    else:
        raise TypeError("Unsupported type")
    return image.convert("RGB")


def draw_bbox(im: np.ndarray, bbox: torch.Tensor, color: TColor) -> np.ndarray:
    """
    Draws a single bounding box on the image.
    If the elements of the bbox are NaNs, we will draw bbox around the whole image.

    Args:
        im: Image
        bbox: Single bounding in the format of [x1, y1, x2, y2]
        color: Tuple of 3 ints
    """
    im_ret = im.copy()
    if not any(torch.isnan(bbox)):
        x1, y1, x2, y2 = list(map(int, bbox))
    elif all(torch.isnan(bbox)):
        x1, y1, x2, y2 = 0, 0, im_ret.shape[1], im_ret.shape[0]
    else:
        raise ValueError("BBox can only consist of all NaNs or all numbers.")

    im_avg_sz = (im_ret.shape[0] + im_ret.shape[1]) / 2
    thickness = max(3, int(0.05 * im_avg_sz))
    im_ret = cv2.rectangle(im_ret, (x1, y1), (x2, y2), thickness=thickness, color=color)

    return im_ret


def get_img_with_bbox(im_path: str, bbox: torch.Tensor, color: TColor) -> np.ndarray:
    """
    Reads the image by its name and draws bbox on it.

    Args:
        im_path: Image path
        bbox: Single bounding box in the format of [x1, y1, x2, y2]. It may also be a list of 4 torch("nan").
        color: Tuple of 3 ints from 0 to 255
    """
    img = imread_cv2(im_path)
    img = draw_bbox(img, bbox, color)
    return img


def square_pad(img: np.ndarray) -> np.ndarray:
    return pad(img, min_height=max(img.shape), min_width=max(img.shape), border_mode=0, value=PAD_COLOR)


def try_to_open_image(im_path: Path, f_imread: TImReader) -> Optional[str]:
    try:
        _ = f_imread(im_path)
        return None
    except Exception:
        return str(im_path)


def find_broken_images(images_list: List[Path], f_imread: TImReader, num_processes: int = 10) -> List[str]:
    try_to_open_image_ = partial(try_to_open_image, f_imread=f_imread)

    with mp.Pool(processes=num_processes) as p:
        results = list(tqdm(p.imap(try_to_open_image_, images_list), total=len(images_list)))

    results = list(filter(lambda x: x is not None, results))
    return results


__all__ = [
    "TImage",
    "TImReader",
    "tensor_to_numpy_image",
    "imread_cv2",
    "imread_pillow",
    "draw_bbox",
    "get_img_with_bbox",
    "square_pad",
]
